#!/usr/bin/python

import gobject

import dbus
import dbus.service
import dbus.mainloop.glib

import os
import slip.dbus.service
import subprocess

import augeas

#######
GRUBBY_CMD        = "/sbin/grubby"
KDUMP_CONFIG_FILE = "/etc/kdump.conf"

# augeas paths
GRUB_PATH = "/files/etc/grub.conf/"
FSTAB_PATH = "/files/etc/fstab/"
KDUMP_PATH = "/files/etc/kdump.conf/"

# can be replaced by using booty?
#             bootloader : (config file, kdump offset, kernel path)
BOOTLOADERS = {
  "grub"   : ("/boot/grub/grub.conf", 16, "/boot"),
  "yaboot" : ("/boot/etc/yaboot.conf", 32, "/boot"),
  "elilo"  : ("/boot/efi/EFI/redhat/elilo.conf", 256, "/boot/efi/EFI/redhat")
}

SUPPORTEDFSTYPES = ("ext2", "ext3", "ext4")

# got from kernel/Documentation/devices.txt
SUPPORTED_MAJOR = [ '2', '3', '8', '9', '13', '14', '19', '21', '22', '28',
                    '31', '33', '34', '36', '40', '44', '45', '47', '48', '49',
                    '50', '51', '52', '53', '54', '55', '56', '57', '65', '66',
                    '67', '68', '69', '70', '71', '88', '89', '90', '91', '94',
                    '99', '128', '129', '130', '131', '132', '133', '134',
                    '135', '147', '180' ]

AUTH         = "org.fedoraproject.systemconfig.kdump"
MECHANISM    = "org.fedoraproject.systemconfig.kdump.mechanism"
KDUMP_OBJECT = "/org/fedoraproject/systemconfig/kdump/object"

def hex2mb_float(hex_code):
    """
    convert hex memory values into MB of RAM in float precision
    """
    return float(int(hex_code, 16)) / 1048576

class SystemConfigKdumpObject(slip.dbus.service.Object):
    def __init__ (self, *p, **k):
        super (SystemConfigKdumpObject, self).__init__ (*p, **k)
        self.bootloader = self.set_bootloader()
        self.augeas = augeas.Augeas()

    @slip.dbus.polkit.require_auth (AUTH + ".noauth")
    @dbus.service.method (MECHANISM,
                          in_signature = '', out_signature = '(bi)')
    def gettotalmem (self):
        """ Get total memory on system """
        total_mem = 0
        accurate = True
        for line in open("/proc/iomem").readlines():
            if line.find("RAM") != -1 or line.find("MMCONFIG") != -1:
                hex_ck_start = line.strip().split("-")[0]
                hex_ck_end = line.strip().split("-")[1].split(":")[0].strip()
                total_mem += int(hex_ck_end, 16) - int(hex_ck_start, 16) + 1
        total_mem = total_mem / 1048576
        if total_mem == 0:
            accurate = False
            for line in open("/proc/meminfo").readlines():
                if line.startswith("MemTotal:"):
                    total_mem = int(line.split()[1]) / 1024
        return (accurate, total_mem)

    @slip.dbus.polkit.require_auth (AUTH + ".noauth")
    @dbus.service.method (MECHANISM,
                          in_signature = '', out_signature = 'a{s(ss)}')
    def getfstab (self):
        """ Get local fs entries from /etc/fstab """
        fstab = {}
        for record in self.augeas.match(FSTAB_PATH + "*"):
           if 'comment' not in record:
               fs_type = self.augeas.get(record + "/vfstype")
               if fs_type not in SUPPORTEDFSTYPES:
                   continue
               fs_spec = self.augeas.get(record + "/spec")
               fs_file = self.augeas.get(record + "/file")
               fstab[fs_spec] = (fs_type, fs_file)
        return fstab

    @slip.dbus.polkit.require_auth (AUTH + ".noauth")
    @dbus.service.method (MECHANISM,
                          in_signature = '', out_signature = 'as')
    def getrawdevices (self):
        """ get devices from /proc/partitions """
        devices = []
        for line in open("/proc/partitions").readlines():
            major = line.strip().split(" ")[0]
            if major in SUPPORTED_MAJOR:
                dev = "/dev/%s" % line.strip().rsplit(" ", 1)[1]
                devices.append(dev)
        return devices

    @slip.dbus.polkit.require_auth (AUTH + ".noauth")
    @dbus.service.method (MECHANISM,
                          in_signature = '', out_signature = 's')
    def getdefaultaction (self):
        return self.augeas.get(KDUMP_PATH + "/default[last()]") or ""

    @slip.dbus.polkit.require_auth (AUTH + ".noauth")
    @dbus.service.method (MECHANISM,
                          in_signature = '', out_signature = 's')
    def getpath (self):
        return self.augeas.get(KDUMP_PATH + "/path[last()]") or ""

    @slip.dbus.polkit.require_auth (AUTH + ".noauth")
    @dbus.service.method (MECHANISM,
                          in_signature = '', out_signature = 's')
    def getcollector (self):
        return self.augeas.get(KDUMP_PATH + "/core_collector[last()]") or ""

    @slip.dbus.polkit.require_auth (AUTH + ".noauth")
    @dbus.service.method (MECHANISM,
                          in_signature = '', out_signature = '(ss)')
    def gettarget (self):
        try:
            target_type = self.augeas.match(
                KDUMP_PATH + "/target[last()]/*")[0]
            target_dest = self.augeas.get(target_type)
            return(target_type.rsplit("/", 1)[-1], target_dest)
        except IndexError:
            # Index error -> no record for target -> default target
            return("", "")


    @slip.dbus.polkit.require_auth (AUTH + ".auth")
    @dbus.service.method (MECHANISM,
                          in_signature = '', out_signature = 'b')
    def unsettarget (self):
        self.augeas.remove(KDUMP_PATH + "target")
        return True

    @slip.dbus.polkit.require_auth (AUTH + ".auth")
    @dbus.service.method (MECHANISM,
                          in_signature = 's', out_signature = 'b')
    def unsetcrashkernel (self, kernel):
        self.augeas.remove(str(
            GRUB_PATH+"*[kernel='%s']/kernel/crashkernel" %kernel))
        return True



    @slip.dbus.polkit.require_auth (AUTH + ".auth")
    @dbus.service.method (MECHANISM,
                          in_signature = 's', out_signature = 'b')
    def setrawdevice (self, devname):
        self.augeas.remove(KDUMP_PATH + "target")
        print "setting target to raw device: ", devname
        self.augeas.set(KDUMP_PATH + "target/raw", str(devname))
        return True

    @slip.dbus.polkit.require_auth (AUTH + ".auth")
    @dbus.service.method (MECHANISM,
                          in_signature = 's', out_signature = 'b')
    def setpath (self, path):
        self.augeas.remove(KDUMP_PATH + "path")
        self.augeas.set(KDUMP_PATH + "path", str(path))
        return True

    @slip.dbus.polkit.require_auth (AUTH + ".auth")
    @dbus.service.method (MECHANISM,
                          in_signature = 's', out_signature = 'b')
    def setnfs (self, nfs):
        self.augeas.remove(KDUMP_PATH + "target")
        self.augeas.set(KDUMP_PATH + "target/net", str(nfs))
        return True

    @slip.dbus.polkit.require_auth (AUTH + ".auth")
    @dbus.service.method (MECHANISM,
                          in_signature = '(ss)', out_signature = 'b')
    def setssh (self, (user, server)):
        self.augeas.remove(KDUMP_PATH + "target")
        self.augeas.set(KDUMP_PATH + "target/net", "%s@%s" %(str(user), str(server)))
        return True

    @slip.dbus.polkit.require_auth (AUTH + ".auth")
    @dbus.service.method (MECHANISM,
                          in_signature = '(ss)', out_signature = 'b')
    def setlocal (self, (fs, partition)):
        self.augeas.remove(KDUMP_PATH + "target")
        self.augeas.set(KDUMP_PATH + "target/%s" % str(fs), str(partition))
        return True


    @slip.dbus.polkit.require_auth (AUTH + ".auth")
    @dbus.service.method (MECHANISM,
                          in_signature = 's', out_signature = 'b')
    def setcollector (self, collector):
        self.augeas.remove(KDUMP_PATH + "core_collector")
        self.augeas.set(KDUMP_PATH + "core_collector", str(collector))
        return True

    @slip.dbus.polkit.require_auth (AUTH + ".auth")
    @dbus.service.method (MECHANISM,
                          in_signature = 's', out_signature = 'b')
    def setdefaultaction (self, action):
        self.augeas.remove(KDUMP_PATH + "default")
        self.augeas.set(KDUMP_PATH + "default", str(action))
        return True

    @slip.dbus.polkit.require_auth (AUTH + ".auth")
    @dbus.service.method (MECHANISM,
                          in_signature = '', out_signature = 'b')
    def unsetdefault (self):
        self.augeas.remove(KDUMP_PATH + "default")
        return True


    @slip.dbus.polkit.require_auth (AUTH + ".auth")
    @dbus.service.method (MECHANISM,
                          in_signature = '', out_signature = 'b')
    def save (self):
        print "Saving it"
        self.augeas.save()
        return True



    @slip.dbus.polkit.require_auth (AUTH + ".noauth")
    @dbus.service.method (MECHANISM,
                          in_signature = '', out_signature = '(ii)')
    def getkdumpmem (self):
        """ Return how much memory is reserved for kdump """
        kdump_mem = 0
        kdump_offset = 0
        for line in open("/proc/iomem").readlines():
            if line.find("Crash kernel") != -1:
                ck_start = int(line.strip().split("-")[0], 16)
                ck_end = int(
                    line.strip().split("-")[1].split(":")[0].strip(),
                    16) + 1
                kdump_mem = (ck_end - ck_start) / 1048576
                kdump_offset = ck_start / 1048576
                break
        if kdump_mem == 0:
            cmdline = open("/proc/cmdline").read()
            if cmdline.find("crashkernel=") > -1:
                crash_string = filter(lambda t: t.startswith("crashkernel="),
                    cmdline.split())[0].split("=")[1]
                tokens = crash_string.split("@")
                kdump_mem = int(tokens[0][:-1])
                if len(tokens) == 2:
                    kdump_offset = int(tokens[1][:-1])
        return (kdump_mem, kdump_offset)


    @slip.dbus.polkit.require_auth (AUTH + ".noauth")
    @dbus.service.method (MECHANISM,
                          in_signature = '', out_signature = 's')
    def getarch (self):
        """ Get architecture we are running on"""
        (cmd, status, std, err) = self.gtkcall('/bin/uname', '-m')
        return std

    @slip.dbus.polkit.require_auth (AUTH + ".noauth")
    @dbus.service.method (MECHANISM,
                          in_signature = '', out_signature = 's')
    def getrunningkernel (self):
        """ Get kernel we are running on"""
        (cmd, status, std, err) = self.gtkcall('/bin/uname', '-r')
        return std

    @slip.dbus.polkit.require_auth (AUTH + ".noauth")
    @dbus.service.method (MECHANISM,
                          in_signature = '', out_signature = 's')
    def getdefaultkernel (self):
        """ Get default kernel name from grubby """
        def_num = int(self.augeas.get(GRUB_PATH + "default")) + 1
        return self.augeas.get(GRUB_PATH + "title[%d]/kernel" % def_num) or ""


    @slip.dbus.polkit.require_auth (AUTH + ".noauth")
    @dbus.service.method (MECHANISM,
                          in_signature = 's', out_signature = 'a{sas}')
    def getcmdline (self, kernel):
        """ Get command line arguments for kernel from grub """
        arg_dict = {}
        title = self.augeas.match("/files/etc/grub.conf/*[kernel = '%s']"
            %str(kernel))[0]
        for arg in self.augeas.match(title + "/kernel/*"):
            value = self.augeas.get(arg) or ""
            arg = (arg, arg.rsplit("[", 1)[0])[arg[-1] == ']']
            arg_name = arg.rsplit("/")[-1]
            if arg_name in arg_dict:
                arg_dict[arg_name].append(value)
            else:
                arg_dict[arg_name] = [value]
        return arg_dict

    @slip.dbus.polkit.require_auth (AUTH + ".auth")
    @dbus.service.method (MECHANISM,
                          in_signature = 'sa{sas}', out_signature = 'b')
    def setkernelargs (self, kernel, arg_dict):
        """ Set command line arguments for kernel  """
        title = self.augeas.match("/files/etc/grub.conf/*[kernel = '%s']"
            %str(kernel))[0]
        self.augeas.remove(title + "/kernel/*")
        for key in arg_dict:
            for n, v in enumerate(arg_dict[key], 1):
                v = (str(v), None)[v==""]
                self.augeas.set(str(title + "/kernel/" + key + "[%d]" %n), v)
        return True

    @slip.dbus.polkit.require_auth (AUTH + ".noauth")
    @dbus.service.method (MECHANISM,
                          in_signature = '', out_signature = 'as')
    def getallkernels (self):
        """ Get all kernel names from grubby """
        arr=[]
        for a in self.augeas.match(GRUB_PATH+"title[*]/kernel"):
            arr.append(self.augeas.get(a))
        return arr

    @slip.dbus.polkit.require_auth (AUTH + ".auth")
    @dbus.service.method (MECHANISM,
                          in_signature = 's', out_signature = '(is)')
    def writedumpconfig (self, config_string):
        """ Write kdump configuration to /etc/kdump.conf
            and return what we write into kdump config file """
        try:
            fd = open(KDUMP_CONFIG_FILE, "w")
            fd.write(config_string)
            fd.close()
        except IOError,(errno, strerror):
            return (errno, strerror)

        # re-read
        return (0, open(KDUMP_CONFIG_FILE).read())

    @slip.dbus.polkit.require_auth (AUTH + ".auth")
    @dbus.service.method (MECHANISM,
                          in_signature = 's', out_signature = '(siss)')
    def writebootconfig (self, config_string):
        """
        Write bootloader configuration.
        in config_string are arguments for grubby divided by `;'
        """
        return self.gtkcall(*([GRUBBY_CMD] + ["--" + self.bootloader] +
                            config_string.split(";")))

    def set_bootloader(self):
        """ Choose which bootloader is on the system """
        bootloader = None
        for (name, (conf, offset, kpath)) in BOOTLOADERS.items():
            # I hope order doesn't matter
            if os.access(conf, os.W_OK):
                bootloader = name
        if bootloader is None:
            return ""
        return bootloader

    @slip.dbus.polkit.require_auth (AUTH + ".auth")
    @dbus.service.method (MECHANISM,
                          in_signature = '(ss)', out_signature = '(siss)')
    def handledumpservice (self, (chkconfig_status, service_op)):
        """ Turn on/off kdump initscript. Start/stop kdump service """
        status = 0
        std = ""
        err = ""
        (cmd, status, std, err) = self.gtkcall("/sbin/chkconfig", "kdump",
                                          chkconfig_status)
        if status > 0:
            return (cmd, status, std, err)

        if service_op != "":
            (cmd, status, std, err) = self.gtkcall("/sbin/service", "kdump",
                                              service_op)
        if status > 0:
            return (cmd, status, std, err)

        if self.bootloader == 'yaboot':
            (cmd, status, std, err) = self.gtkcall('/sbin/ybin')
            if status > 0:
                return (cmd, status, std, err)

        return (cmd, status, std, err)

    @slip.dbus.polkit.require_auth (AUTH + ".auth")
    @dbus.service.method (MECHANISM,
                          in_signature = '', out_signature = '(siss)')
    def getservicestatus (self):
        """ Get current status of the kdump service """
        return self.gtkcall("/sbin/service", "kdump", "status")

    def gtkcall (self, *args):
        """
        Call command args[0] with args arguments
        Return the tuple with (called command, retcode, stdout, stderr)
        """
        the_call = subprocess.Popen(args, stdout = subprocess.PIPE,
                                    stderr = subprocess.PIPE)
        stdout, stderr = the_call.communicate()
        print "subprocess call:"
        print args
        if stdout:
            print "-> stdout = " + stdout
        if stderr:
            print "-> stderr = " + stderr
        return ("".join(str(a) + " " for a in args),
            the_call.returncode, stdout or "", stderr or "")

if __name__ == '__main__':
    dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)

    BUS = dbus.SystemBus ()

    NAME = dbus.service.BusName (MECHANISM, BUS)
    OBJECT = SystemConfigKdumpObject (BUS, KDUMP_OBJECT)

    MAINLOOP = gobject.MainLoop ()
    slip.dbus.service.set_mainloop (MAINLOOP)
    MAINLOOP.run ()
